@@ -1,8 +1,10 @@
 from __future__ import annotations
 
+import base64
 import os
 import os.path
-from typing import TYPE_CHECKING, Any, Dict, List, Optional, cast
+from time import sleep
+from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, cast
 
 import boto3
 from typing_extensions import TypedDict
@@ -12,17 +14,17 @@ from aec.util.errors import NoInstancesError
 from aec.util.threads import executor
 
 if TYPE_CHECKING:
+    from mypy_boto3_ec2.literals import InstanceTypeType
     from mypy_boto3_ec2.type_defs import (
         BlockDeviceMappingTypeDef,
         FilterTypeDef,
-        TagTypeDef,
         InstanceStatusSummaryTypeDef,
         TagSpecificationTypeDef,
+        TagTypeDef,
     )
-    from mypy_boto3_ec2.literals import InstanceTypeType
 
 import aec.command.ami as ami_cmd
-import aec.util.ec2 as util_tags
+import aec.util.tags as util_tags
 from aec.util.config import Config
 from aec.util.ec2_types import RunArgs
 
@@ -50,9 +52,12 @@ def launch(
     instance_type: Optional[str] = None,
     key_name: Optional[str] = None,
     userdata: Optional[str] = None,
+    wait_ssm: bool = False,
 ) -> List[Instance]:
     """Launch a tagged EC2 instance with an EBS volume."""
 
+    template = template or config.get("launch_template", None)
+
     if not (template or ami):
         raise ValueError("Please specify either an ami or a launch template")
 
@@ -106,7 +111,8 @@ def launch(
     elif not template:
         print(
             "WARNING: You have not specified a key pair.",
-            "You will only be able to connect to this instance if it is configured for EC2 Instance Connect or Systems Manager Session Manager.",
+            "You will only be able to connect to this instance if it is configured for "
+            + "EC2 Instance Connect or Systems Manager Session Manager.",
         )
 
     if instance_type:
@@ -148,26 +154,53 @@ def launch(
         runargs["IamInstanceProfile"] = {"Arn": iam_instance_profile_arn}
 
     if userdata:
-        runargs["UserData"] = read_file(userdata)
+        if userdata.startswith("http://") or userdata.startswith("https://"):
+            runargs["UserData"] = read_url(userdata)
+        else:
+            runargs["UserData"] = read_file(userdata)
 
-    # use IMDSv2 to prevent SSRF
-    runargs["MetadataOptions"] = {"HttpTokens": "required"}
+    # use IMDSv2 to prevent SSRF and enable instance metadata tags
+    runargs["MetadataOptions"] = {"HttpTokens": "required", "InstanceMetadataTags": "enabled"}
 
     region_name = ec2_client.meta.region_name
 
-    print(f"Launching a {instance_type} in {region_name}{vpc_name} named {name} using {desc} ... ")
+    print(
+        f"Launching {'a ' + instance_type if instance_type else 'an instance'} in "
+        + f"{region_name}{vpc_name} named {name} using {desc} ... "
+    )
     response = ec2_client.run_instances(**runargs)
 
-    instance = response["Instances"][0]
+    instance_id = response["Instances"][0]["InstanceId"]
 
     waiter = ec2_client.get_waiter("instance_running")
-    waiter.wait(InstanceIds=[instance["InstanceId"]])
+    waiter.wait(InstanceIds=[instance_id])
 
-    # TODO: wait until instance checks passed (as they do in the console)
+    if wait_ssm:
+        print(f"Instance {instance_id} running. Waiting for SSM agent to come online ...")
+        _wait_ssm_agent_online(config, [instance_id])
 
     # the response from run_instances above always contains an empty string
     # for PublicDnsName, so we call describe to get it
-    return describe(config=config, name=name)
+    return describe(config=config, name=instance_id)
+
+
+def _wait_ssm_agent_online(config: Config, instance_ids: List[str]) -> None:
+    """
+    Wait for ssm to come online.
+
+    This ensures the instance is ready to accept ssh logins.
+    """
+    #
+    client = boto3.client("ssm", region_name=config.get("region", None))
+
+    timeout = 60 * 3  # seconds
+    for _ in range(timeout):
+        response = client.describe_instance_information(Filters=[{"Key": "InstanceIds", "Values": instance_ids}])
+        if response["InstanceInformationList"]:
+            return
+        sleep(1)
+
+    raise TimeoutError(f"SSM agent not online after {timeout} seconds")
 
 
 def describe(
@@ -177,7 +210,7 @@ def describe(
     include_terminated: bool = False,
     show_running_only: bool = False,
     sort_by: str = "State,Name",
-    columns: Optional[str] = None,
+    columns: str = "InstanceId,State,Name,Type,DnsName,LaunchTime,ImageId",
 ) -> List[Instance]:
     """List EC2 instances in the region."""
 
@@ -187,34 +220,46 @@ def describe(
     if show_running_only:
         filters.append({"Name": "instance-state-name", "Values": ["pending", "running"]})
 
-    response = ec2_client.describe_instances(Filters=filters)
+    kwargs: Dict[str, Any] = {"MaxResults": 1000, "Filters": filters}
 
-    # print(response["Reservations"][0]["Instances"][0])
+    response = ec2_client.describe_instances(**kwargs)
 
-    cols = columns.split(",") if columns else ["InstanceId", "State", "Name", "Type", "DnsName"]
+    # import json; print(json.dumps(response))
+
+    cols = columns.split(",")
 
     # don't sort by cols we aren't showing
     sort_cols = [sc for sc in sort_by.split(",") if sc in cols]
 
     instances: List[Instance] = []
-    for r in response["Reservations"]:
-        for i in r["Instances"]:
-            if include_terminated or i["State"]["Name"] != "terminated":
-                desc: Instance = {}
+    while True:
+        for r in response["Reservations"]:
+            for i in r["Instances"]:
+                if include_terminated or i["State"]["Name"] != "terminated":
+                    desc: Instance = {}
 
-                for col in cols:
-                    if col == "State":
-                        desc[col] = i["State"]["Name"]
-                    elif col == "Name":
-                        desc[col] = util_tags.get_value(i, "Name")
-                    elif col == "Type":
-                        desc[col] = i["InstanceType"]
-                    elif col == "DnsName":
-                        desc[col] = i["PublicDnsName"] if i.get("PublicDnsName", None) != "" else i["PrivateDnsName"]
-                    else:
-                        desc[col] = i.get(col, None)
+                    for col in cols:
+                        if col == "State":
+                            desc[col] = i["State"]["Name"]
+                        elif col == "Name":
+                            desc[col] = util_tags.get_value(i, "Name")
+                        elif col == "Type":
+                            desc[col] = i["InstanceType"]
+                        elif col == "DnsName":
+                            desc[col] = (
+                                i["PublicDnsName"] if i.get("PublicDnsName", None) != "" else i["PrivateDnsName"]
+                            )
+                        else:
+                            desc[col] = i.get(col, None)
 
-                instances.append(desc)
+                    instances.append(desc)
+
+        next_token = response.get("NextToken", None)
+        if next_token:
+            kwargs["NextToken"] = next_token
+            response = ec2_client.describe_instances(**kwargs)
+        else:
+            break
 
     return sorted(
         instances,
@@ -226,7 +271,7 @@ def describe_tags(
     config: Config,
     name: Optional[str] = None,
     name_match: Optional[str] = None,
-    keys: List[str] = [],
+    keys: Sequence[str] = [],
     volumes: bool = False,
 ) -> List[Dict[str, Any]]:
     """List EC2 instances or volumes with their tags."""
@@ -267,7 +312,7 @@ def tag(
 
 
 def instance_tags(
-    config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: List[str] = []
+    config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: Sequence[str] = []
 ) -> List[Dict[str, Any]]:
     """List EC2 instances with their tags."""
 
@@ -292,7 +337,7 @@ def instance_tags(
 
 
 def volume_tags(
-    config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: List[str] = []
+    config: Config, name: Optional[str] = None, name_match: Optional[str] = None, keys: Sequence[str] = []
 ) -> List[Dict[str, Any]]:
     """List EC2 volumes with their tags."""
 
@@ -314,7 +359,11 @@ def volume_tags(
     return sorted(volumes, key=lambda i: str(i["Name"]))
 
 
-def start(config: Config, name: str) -> List[Instance]:
+def start(
+    config: Config,
+    name: str,
+    wait_ssm: bool = False,
+) -> List[Instance]:
     """Start EC2 instance."""
 
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
@@ -336,6 +385,10 @@ def start(config: Config, name: str) -> List[Instance]:
     waiter = ec2_client.get_waiter("instance_running")
     waiter.wait(InstanceIds=instance_ids)
 
+    if wait_ssm:
+        print(f"Instance {', '.join(instance_ids)} running. Waiting for SSM agent to come online ...")
+        _wait_ssm_agent_online(config, instance_ids)
+
     return describe(config, name)
 
 
@@ -447,16 +500,30 @@ def templates(config: Config) -> List[Dict[str, Any]]:
     ]
 
 
-def status(config: Config) -> List[Dict[str, Any]]:
+def status(
+    config: Config,
+    name: Optional[str] = None,
+    name_match: Optional[str] = None,
+) -> List[Dict[str, Any]]:
     """Describe instances status checks."""
     ec2_client = boto3.client("ec2", region_name=config.get("region", None))
 
-    kwargs = {"MaxResults": 50}
+    kwargs: Dict[str, Any] = {"MaxResults": 1000}
 
     response_fut = executor.submit(ec2_client.describe_instance_status, **kwargs)
     instances = executor.submit(describe_running_instances_names, config).result()
     response = response_fut.result()
 
+    def match(instance_id: str, instance_name: Optional[str]) -> bool:
+        # describe_instance_status doesn't support name filters in the request so match here
+        if not name and not name_match:
+            return True
+
+        if name is not None:
+            return (name.startswith("i-") and name == instance_id) or (name == instance_name)
+
+        return not name_match or name_match in (instance_name or "")
+
     statuses = []
     while True:
         statuses.extend(
@@ -469,12 +536,13 @@ def status(config: Config) -> List[Dict[str, Any]]:
                     "Instance status check": status_text(i["InstanceStatus"]),
                 }
                 for i in response["InstanceStatuses"]
+                if match(i["InstanceId"], instances.get(i["InstanceId"], None))
             ]
         )
 
         next_token = response.get("NextToken", None)
         if next_token:
-            kwargs = {"NextToken": next_token}
+            kwargs["NextToken"] = next_token
             response = ec2_client.describe_instance_status(**kwargs)
         else:
             break
@@ -492,11 +560,33 @@ def status_text(summary: InstanceStatusSummaryTypeDef, key: str = "reachability"
     )
 
 
-def name_filters(name: Optional[str] = None, name_match: Optional[str] = None) -> List[FilterTypeDef]:
-    if name and name.startswith("i-"):
-        return [{"Name": "instance-id", "Values": [name]}]
-    elif name:
-        return [{"Name": "tag:Name", "Values": [name]}]
+def user_data(config: Config, name: str) -> Optional[str]:
+    """Describe user data for an instance."""
+    ec2_client = boto3.client("ec2", region_name=config.get("region", None))
+
+    if not name:
+        # avoid describing all instances when there's no name
+        raise NoInstancesError(name=name)
+
+    instances = describe(config, name)
+
+    if not instances:
+        raise NoInstancesError(name=name)
+
+    instance_id = instances[0]["InstanceId"]
+    response = ec2_client.describe_instance_attribute(Attribute="userData", InstanceId=instance_id)
+
+    try:
+        return base64.b64decode(response["UserData"]["Value"]).decode("utf-8")
+    except KeyError:
+        return None
+
+
+def name_filters(name_or_id: Optional[str] = None, name_match: Optional[str] = None) -> List[FilterTypeDef]:
+    if name_or_id and name_or_id.startswith("i-"):
+        return [{"Name": "instance-id", "Values": [name_or_id]}]
+    elif name_or_id:
+        return [{"Name": "tag:Name", "Values": [name_or_id]}]
     elif name_match:
         return [{"Name": "tag:Name", "Values": [f"*{name_match}*"]}]
     else:
@@ -506,3 +596,11 @@ def name_filters(name: Optional[str] = None, name_match: Optional[str] = None) -
 def read_file(filepath: str) -> str:
     with open(os.path.expanduser(filepath)) as file:
         return file.read()
+
+
+def read_url(url: str) -> str:
+    import requests
+
+    r = requests.get(url)
+    r.raise_for_status()
+    return r.text
