@@ -13,7 +13,7 @@ import aec.util.cli as cli
 import aec.util.config as config
 import aec.util.configure as configure
 import aec.util.display as display
-from aec.util.cli import Arg, Cmd
+from aec.util.cli import Arg, Cmd, parameter_defaults
 from aec.util.errors import HandledError
 
 config_arg = Arg("--config", help="Section of the config file to use")
@@ -52,8 +52,8 @@ ec2_cli = [
         Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
         Arg("-r", "--show-running-only", action='store_true', help="Show running or pending instances only"),
         Arg("-it", "--include-terminated", action='store_true', help="Include terminated instances"),
-        Arg("-s", "--sort-by", type=str, help="Sort by one or more fields", default="State,Name"),
-        Arg("-c", "--columns", type=str, help="Customise the columns shown", default="InstanceId,State,Name,Type,DnsName,LaunchTime,ImageId"),
+        Arg("-s", "--sort-by", type=str, help="Sort by one or more fields", default=parameter_defaults(ec2.describe)["sort_by"]),
+        Arg("-c", "--columns", type=str, help="Customise the columns shown", default=parameter_defaults(ec2.describe)["columns"]),
     ]),
     Cmd(ec2.launch, [
         config_arg,
@@ -64,7 +64,8 @@ ec2_cli = [
         Arg("--encrypted", type=bool, help="Whether the EBS volume is encrypted", default=True),
         Arg("--instance-type", type=str, help="Instance type"),
         Arg("--key-name", type=str, help="Key name"),
-        Arg("--userdata", type=str, help="Path to user data file")
+        Arg("--userdata", type=str, help="User data file path or http URL"),
+        Arg("-w", "--wait-ssm", action='store_true', help="Wait until the SSM agent is online before exiting"),
     ]),
     Cmd(ec2.logs, [
         config_arg,
@@ -77,7 +78,8 @@ ec2_cli = [
     ]),
     Cmd(ec2.start, [
         config_arg,
-        Arg("name", type=str, help="Name tag of instance or instance id")
+        Arg("name", type=str, help="Name tag of instance or instance id"),
+        Arg("-w", "--wait-ssm", action='store_true', help="Wait until the SSM agent is online before exiting"),
     ]),
     Cmd(ec2.stop, [
         config_arg,
@@ -97,7 +99,9 @@ ec2_cli = [
         Arg("-k", "--keys", type=str, nargs='*', metavar="KEY", help="Tags to display", default = []),
     ], name = "tags"),
     Cmd(ec2.status, [
-        config_arg
+        config_arg,
+        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
+        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
     ]),
     Cmd(ec2.templates, [
         config_arg
@@ -105,7 +109,11 @@ ec2_cli = [
     Cmd(ec2.terminate, [
         config_arg,
         Arg("name", type=str, help="Name tag of instance or instance id")
-    ])
+    ]),
+    Cmd(ec2.user_data, [
+        config_arg,
+        Arg("name", type=str, help="Name tag of instance or instance id"),
+    ]),
 ]
 
 ami_cli = [
@@ -115,11 +123,18 @@ ami_cli = [
     ]),
     Cmd(ami.describe, [
         config_arg,
-        Arg("--ami", type=str, help="Filter to this AMI id"),
+        Arg("id", type=str, nargs='?', help="Filter to this AMI id"),
         Arg("--owner", type=str, help="Filter to this owning account"),
         Arg("-q", type=str, dest='name_match', help="Filter to images with a name containing NAME_MATCH."),
         Arg("--show-snapshot-id", action='store_true', help="Show snapshot id")
     ]),
+    Cmd(ami.describe_tags, [
+        config_arg,
+        Arg("id", type=str, nargs='?', help="Filter to this AMI id"),
+        Arg("--owner", type=str, help="Filter to this owning account"),
+        Arg("-q", type=str, dest='name_match', help="Filter to images with a name containing NAME_MATCH."),
+        Arg("-k", "--keys", type=str, nargs='*', metavar="KEY", help="Tags to display", default = []),
+    ], name = "tags"),
     Cmd(ami.share, [
         config_arg,
         Arg("ami", type=str, help="AMI id"),
@@ -142,7 +157,9 @@ ssm_cli = [
         config_arg
     ]),
     Cmd(ssm.describe, [
-        config_arg
+        config_arg,
+        Arg("name", type=str, nargs='?', help="Filter to instances with this Name tag or instance id."),
+        Arg("-q", type=str, dest='name_match', help="Filter to instances with a Name tag containing NAME_MATCH."),
     ]),
     Cmd(ssm.invocations, [
         config_arg,
@@ -172,7 +189,8 @@ ssm_cli = [
 
 
 def build_parser() -> argparse.ArgumentParser:
-    parser = argparse.ArgumentParser(description="aws ec2 cli")
+    # prog=aec so that we run in cog the program name isn't cog
+    parser = argparse.ArgumentParser(prog="aec", description="aws ec2 cli")
     subparsers = parser.add_subparsers(title="commands")
 
     cli.add_command_group(subparsers, "configure", "Configure subcommands", configure_cli)
